package h06.problems;

import h06.ui.DrawInstruction;
import org.tudalgo.algoutils.student.annotation.DoNotTouch;
import org.tudalgo.algoutils.student.annotation.StudentImplementationRequired;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static org.tudalgo.algoutils.student.Student.crash;

/**
 * A class to generate draw instructions in order to draw a dragon curve.
 *
 * @author Manuel Peters
 */
public class Fractals {

    /**
     * Default Constructor for this class.
     */
    @DoNotTouch
    public Fractals() {}

    /**
     * This method calculates a raised to the power of b using recursion.
     * a and b are expected to be non-negative integers.
     *
     * @param a the base, must be non-negative
     * @param b the exponent, must be non-negative
     * @return the result of a raised to the power of b
     */
    @StudentImplementationRequired
    public static int pow(int a, int b) {
        return (int) Math.pow(a, b);
    }

    /**
     * This method combines two arrays of DrawInstruction objects into a single array.
     * The elements of the first array are followed by the elements of the second array in the new array.
     *
     * @param arr1 the first array of type DrawInstruction
     * @param arr2 the second array of type DrawInstruction
     * @return A new array containing all elements of arr1 followed by all elements of arr2
     */
    @StudentImplementationRequired
    public static DrawInstruction[] concatenate(DrawInstruction[] arr1, DrawInstruction[] arr2) {
        do {
            return Stream.of(arr1, arr2).flatMap(Stream::of).toArray(DrawInstruction[]::new); // TODO: H3.2 - remove if implemented
        } while(false);
    }

    /**
     * This method creates a new array that is a copy of the input array arr, but with the element at the specified
     * index idx replaced by elem.
     *
     * @param arr the original array of type DrawInstruction
     * @param idx the index at which to replace the element
     * @param elem the new DrawInstruction to place at the specified index
     * @return A new array with the element at idx replaced by elem
     */
    @StudentImplementationRequired
    public static DrawInstruction[] replaceAtIndex(DrawInstruction[] arr, int idx, DrawInstruction elem) {
        do {
            List<DrawInstruction> x = new ArrayList<>(List.of(arr));
            x.set(idx, elem);
            return x.toArray(DrawInstruction[]::new);
        } while (true);
    }

    static String fracToString(DrawInstruction[] frac) {
        return Stream.of(frac).map(String::valueOf).collect(Collectors.joining());
    }

    static DrawInstruction[] stringToFrac(String string) {
        return string.chars()
            .mapToObj(c -> (char) c)
            .map(String::valueOf)
            .map(x -> switch (x) {
                case "D": yield DrawInstruction.DRAW_LINE;
                case "R": yield DrawInstruction.TURN_RIGHT;
                case "L": yield DrawInstruction.TURN_LEFT;
                default:
                    throw new IllegalStateException("Unexpected value: " + x);
            })
            .toArray(DrawInstruction[]::new);
    }

    /**
     * Generates an array of DrawInstruction objects to draw a dragon curve of order n
     *
     * @param n The order of the dragon curve to generate
     * @return an array of DrawInstruction objects to draw a dragon curve of order n
     */
    @StudentImplementationRequired
    public static DrawInstruction[] dragonCurve(int n) {
        return stringToFrac(dragonCurveString(n));
    }
    public static String dragonCurveString(int n) {
        if (n < 0) return "D";

        return switch (n) {
            case 0: yield "D";
            case 1: yield "DRD";
            default: yield ((dragonCurveString(n-1) + "R") + stringReplace(dragonCurveString(n-1), pow(2, n-1)-1, "L"));
        };
    }

    static String stringReplace(String a, int n, String b) {
        return a.substring(0, n) + b + a.substring(n+1);
    }


    public static <A> List<A> replaceAll(List<A> list, A a, List<A> newThings) {
        return list.stream().flatMap(x -> a.equals(x) ? newThings.stream() : Stream.of(x)).toList();
    }

    /**
     * Generates an array of DrawInstruction objects to draw a koch snowflake of order n
     *
     * @param n The order of the koch snowflake to generate
     * @return an array of DrawInstruction objects to draw a koch snowflake of order n
     */
    @StudentImplementationRequired
    public static DrawInstruction[] kochSnowflake(int n) {
        return stringToFrac(kockSnowflakeString(n));
    }

    public static String kockSnowflakeString(int n) {
        if (n < 0) return "DRRDRRD";

        return switch (n) {
            case 0: yield "DRRDRRD";
            default: yield kockSnowflakeString(n-1).replaceAll("D", "DLDRRDLD");
        };
    }


}
